(*
 Example of explaining away from
 http://www.ai.mit.edu/~murphyk/Bayes/bnintro.html#explainaway

 Suppose you have to be brainy or sporty to get into college.
 B S P(C=1) P(C=2)  1=false 2=true
 1 1 1.0    0.0 
 2 1 0.0    1.0
 1 2 0.0    1.0
 2 2 0.0    1.0


 If we observe that you are in college, you must be either brainy or
 sporty or both.  If we observe you are in college and sporty, it is
 less likely you are brainy, since brainy-ness and sporty-ness compete
 as causal explanations of the effect.

 B  S
  \/
  C
*)

functor Brainy() = 
struct
open GM
structure Eng : INFERENCE = NaiveInfer
structure IP = IPotential

val [FALSE, TRUE] = List.map RVar.intToValue [0,1] 
val [B,S,C] = List.map RVar.newVar [2,2,2]
val _ = RVar.giveName B "Brainy"
val _ = RVar.giveName S "Sporty"
val _ = RVar.giveName C "College"

val cpds = List.map CPD.make
		    [([0.5, 0.5], B, []),
		     ([0.5, 0.5], S, []),
		     ([1.0, 0.0, 0.0, 1.0,
		       0.0, 1.0, 0.0, 1.0], C, [B,S])]
val local_factors = List.map LocalFactor.CPD cpds
val gmodel = GModel.make local_factors

(* Pop up a display showing the model *)
val _ = Util.showDot (GModel.toDotString gmodel)

val engine = Eng.makeEngine (gmodel, [], [])

val engine = Eng.enterEvidence engine (Eng.Hard (C, TRUE))
val PB = Eng.query engine [B] 
val _ = Util.assert (Util.approxEq 10E~4 (IP.index PB [TRUE]) 0.66666) 

val engine = Eng.enterEvidence engine (Eng.Hard (S, TRUE))
val PB' = Eng.query engine [B]
val _ = Util.assert (Util.approxEq 10E~4 (IP.index PB' [TRUE]) 0.5) 

end
