(* Lawn sprinker example from Russell and Norvig p454 *)
functor Sprinkler() =
struct
open GM
structure L : PARAM_LEARNER = ParamLearn
structure IP = IPotential

(* Setup the Random Variables *)
val [FALSE, TRUE] = List.map RVar.intToValue [0,1] 
val [C,S,R,W] = List.map RVar.newVar [2,2,2,2]
val _ = RVar.giveName C "Cloudy"
val _ = RVar.giveName S "Sprinkler"
val _ = RVar.giveName R "Raining"
val _ = RVar.giveName W "Wet Grass"

(* Create the local distributions *)
val cpds = List.map CPD.make [ ([0.5, 0.5], C, []),
			       ([0.8, 0.2, 0.2, 0.8], R, [C]),
			       ([0.5, 0.5, 0.9, 0.1], S, [C]),
			       ([1.0, 0.0, 0.1, 0.9, 
				 0.1, 0.9, 0.01, 0.99], 
				W, [S,R]) ]
val local_factors = List.map LocalFactor.CPD cpds

(* Make the model *)
val gmodel = GModel.make local_factors

(* Pop up a display showing the model for fun*)
val _ = Util.showDot (GModel.toDotString gmodel)

(* Perform learning in the model *)
val learner = L.makeLearner (gmodel, [],[] )

(* Genorate training data *)

end
