(* Lawn sprinker example from Russell and Norvig p454 *)
functor Sprinkler() =
struct
open GM
structure Eng : INFERENCE = NaiveInfer
structure IP = IPotential

(* Setup the Random Variables *)
val [FALSE, TRUE] = List.map RVar.intToValue [0,1] 
val [C,S,R,W] = List.map RVar.newVar [2,2,2,2]
val _ = RVar.giveName C "Cloudy"
val _ = RVar.giveName S "Sprinkler"
val _ = RVar.giveName R "Raining"
val _ = RVar.giveName W "Wet Grass"

(* Create the local distributions *)
val factors = List.map (LocalFactor.CPD o CPD.make) 
		       [ ([0.5, 0.5], C, []),
			 ([0.8, 0.2, 0.2, 0.8], R, [C]),
			 ([0.5, 0.5, 0.9, 0.1], S, [C]),
			 ([1.0, 0.0, 0.1, 0.9, 
			   0.1, 0.9, 0.01, 0.99], W, [S,R]) ]
(* Make the model *)
val gmodel = GModel.make factors

(* Pop up a display showing the model *)
val _ = Util.showDot (GModel.toDotString gmodel)

(* Perform inference in the model *)
val engine = Eng.makeEngine (gmodel, [],[] )

(* Add evidence that the grass is wet *)
val engine = Eng.enterEvidence engine (Eng.Hard (W, TRUE))

val P_S = Eng.query engine [S]
val _ = Util.assert (Util.approxEq 10E~4 (IP.index P_S [TRUE]) 0.4298)
val P_R = Eng.query engine [R]
val _ = Util.assert (Util.approxEq 10E~4 (IP.index P_R [TRUE]) 0.7079)

(* Add extra evidence that its Raining *)
val engine = Eng.enterEvidence engine (Eng.Hard (R, TRUE))
val P_S = Eng.query engine [S]
val _ = Util.assert (Util.approxEq 10E~4 (IP.index P_S [TRUE]) 0.1945)

(* The sprinler is likely to be on due to the rain "explaining away"
   the wet grass *)

end
