(* DGM : Interface for a (completely discrete) graphical model library.

  This is simply a collection of all structures needed to use the GM library 
  for computing over discrete models. This is mostly functor applications.
*)

signature GM =
sig

structure RVar : D_R_VAR
structure IPotential : D_IPOTENTIAL (* This should be hidden eventually *)
structure Potential : D_POTENTIAL 
structure CPD : D_CPD 
structure LocalFactor : D_LOCAL_FACTOR 
structure GModel : D_G_MODEL 

(* Inference Engines *)
structure NaiveInfer : INFERENCE 
structure VarElimInfer : INFERENCE
structure JoinTreeInfer : INFERENCE
structure LazyInfer : INFERENCE

(* Learning *)
structure ParamLearn : PARAM_LEARN

(* The sharing restrictions on the internal structures *)
sharing RVar = IPotential.RVar = Potential.RVar = CPD.RVar = LocalFactor.RVar
	    = GModel.RVar = NaiveInfer.RVar = VarElimInfer.RVar
	    = JoinTreeInfer.RVar = LazyInfer.RVar
and LocalFactor = GModel.LocalFactor
and LocalFactor.CPD = CPD
and LocalFactor.Potential = Potential
and GModel = NaiveInfer.GModel = VarElimInfer.GModel = JoinTreeInfer.GModel
	   = LazyInfer.GModel
and IPotential = NaiveInfer.IP = VarElimInfer.IP = JoinTreeInfer.IP
	       = LazyInfer.IP

(* for command-line access *)
val main : string * string list -> OS.Process.status

end
