(* D_G_MODEL : A discrete graphical model

   Contains methods for building and creating dgms.

*)
signature D_G_MODEL =
sig
    structure LocalFactor : D_LOCAL_FACTOR
    structure IP : D_IPOTENTIAL (* FIXME-- don't want to return IPs*)
    structure RVar : D_R_VAR
    sharing IP.RVar = LocalFactor.RVar = RVar
		      
    structure DGraph : DGRAPH
    structure UGraph : UGRAPH
    sharing UGraph.DGraph = DGraph

    exception Error of string

    (* A discrete graphical model is the untion of all supported models. 
       - Should the guts of these types be exposed? Why not.
     *)
    datatype gmodel = BNet of (IP.ipot list * RVar.var DGraph.dgraph)
		    | MNet of (IP.ipot list * RVar.var UGraph.ugraph)

    val make : LocalFactor.factor list -> gmodel

    val toString : gmodel -> string
    val toDotString : gmodel -> string
    (*val getFactors : gmodel -> factor list*)

    (* Private stuff *)
    val getIPotentials : gmodel -> IP.ipot list
    val addIPotential : gmodel -> IP.ipot -> gmodel

    (* General graph queries*)
end
