functor DGModelFn(structure LocalFactor : D_LOCAL_FACTOR
	       structure IP : D_IPOTENTIAL
	       structure RVar : D_R_VAR
	       sharing IP.RVar = LocalFactor.RVar = RVar
	       and LocalFactor.IPotential = IP) 
	:> D_G_MODEL where LocalFactor = LocalFactor
                  and IP = IP
		  and RVar = RVar =
struct

structure LocalFactor = LocalFactor
structure IP = IP
structure RVar = RVar

structure DGraph = UGraph.DGraph
structure UGraph = UGraph
structure CPD = LocalFactor.CPD
structure Pot = LocalFactor.Potential

structure RVarSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)

exception Error of string

datatype gmodel = BNet of (IP.ipot list * RVar.var DGraph.dgraph)
		| MNet of (IP.ipot list * RVar.var UGraph.ugraph)

fun make flist = 
    if List.all (fn (LocalFactor.CPD _) => true
		  | _  => false) 
			flist
    then makeBNet flist
    else if List.all (fn (LocalFactor.Potential _) => true
		       | _  => false) 
		     flist
    then makeMNet flist
    else raise Error "Graphical model not supported\n"

and makeBNet flist =
        let
	val var_set = List.foldl (fn (f,set)=> 
				     RVarSet.addList (set, LocalFactor.scope f))
		      RVarSet.empty 
		      flist
	val sorted_vars = RVarSet.listItems var_set 
	val ipots = List.map LocalFactor.toIPotential flist
	val graph = DGraph.make sorted_vars
	(* map from a cpd to a child*parent paris of vertex ids *)
	fun cpd2vIds cpd = 
	    let
		val childid = valOf (DGraph.getVertexId graph 
							(CPD.getChild cpd))
		val parentids = List.map (fn x=> valOf (DGraph.getVertexId 
							    graph x))
					 (CPD.getParents cpd)
	    in (childid, parentids) end	 
	val graph = 
	    List.foldr 
		(fn ((LocalFactor.CPD cpd),g) => 
		    DGraph.addParents g (cpd2vIds cpd))
		graph
		flist
    in
	BNet (ipots, graph) 
    end

and makeMNet flist =
        let
	val var_set = List.foldl (fn (f,set)=> 
				     RVarSet.addList (set, LocalFactor.scope f))
		      RVarSet.empty 
		      flist
	val sorted_vars = RVarSet.listItems var_set 
	val ipots = List.map LocalFactor.toIPotential flist
	val graph = UGraph.make sorted_vars

	(* map from a potential to vertex ids *)
	fun pot2vIds pot = List.map (fn x=> valOf (UGraph.getVertexId 
						       graph x))
				    (Pot.getScope pot)
	val graph = 
	    List.foldr
		(fn (LocalFactor.Potential pot, g)
		    => UGraph.connect g (pot2vIds pot))
		graph
		flist
	in
	    MNet (ipots, graph)
	end


fun getIPotentials (BNet (ipots, _)) = ipots
  | getIPotentials (MNet (ipots, _)) = ipots

fun addIPotential (BNet (ipots, g)) ipot = BNet(ipot::ipots, g)
  | addIPotential (MNet (ipots, g)) ipot = MNet(ipot::ipots, g)


fun toString (BNet (ipots, dgraph)) = 
    "BNet : \n" ^ (String.concat (List.map IP.toString ipots))
    ^ "\n" ^ (DGraph.toString dgraph RVar.toString)
  | toString (MNet (ipots, ugraph)) = 
    "MNet : \n" ^ (String.concat (List.map IP.toString ipots))
    ^ "\n" ^ (UGraph.toString ugraph RVar.toString)

fun toDotString (BNet (_, dgraph)) = DGraph.toDotString dgraph RVar.toString
  | toDotString (MNet (_, ugraph)) = UGraph.toDotString ugraph RVar.toString

end
