(* JOIN_TREE_INNARDS : Provide the methods needed to build a jtree. 

   This is used so that the code to build regular and LAZY join trees 
   can be shared. 			       
*)
signature JOIN_TREE_INNARDS =
sig
    structure GModel : D_G_MODEL
    structure IP : D_IPOTENTIAL
    structure RVar : D_R_VAR
    sharing GModel.RVar = IP.RVar = RVar
    and GModel.IP = IP

    type pot

    val scope : pot -> RVar.var list
    val initialClique : RVar.var list -> pot
    val initialSep : RVar.var list -> pot
    val addPotToClique : IP.ipot * pot -> pot
end

(* JOIN_TREE : A Join Tree of potentials *)
signature JOIN_TREE =
sig
    structure JTI : JOIN_TREE_INNARDS
    structure GModel : D_G_MODEL
    structure IP : D_IPOTENTIAL
    structure RVar : D_R_VAR
    sharing GModel.RVar = IP.RVar = RVar
    and GModel.IP = IP
    and JTI.GModel = GModel
    and JTI.IP = IP
    and JTI.RVar = RVar

    type pot (* = JTI.pot*)

    (* A join-tree is represented as a recursive datatype as a clique
      potential and a list of seperator potenials and thier
      corrisponding (sub) jtrees. This makes the root explicit,
      although new roots can be had by a simple rotation. *)
    datatype jtree = Clique of pot * ((pot * jtree) list)

    val size : jtree -> int
    val scope : jtree -> RVar.var list
    val toDotString : jtree -> string

    (* Build an *inconsistent* join tree *)
    val build : GModel.gmodel -> jtree
end

