functor JoinTreeInferFn(structure GModel : D_G_MODEL
			structure RVar : D_R_VAR
			structure IP : D_IPOTENTIAL
			sharing GModel.RVar = RVar 
		       and GModel.IP = IP ) 
	:> INFERENCE where GModel = GModel
                       and IP = IP
		       and RVar = RVar = 
struct

structure GModel = GModel
structure IP = IP
structure RVar = RVar
structure RVSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)
structure JTI : JOIN_TREE_INNARDS = (* transparent ascription *)
struct                              (*  is useful here! *)
structure GModel = GModel
structure IP = IP
structure RVar = RVar

type pot = IP.ipot
val scope = IP.scope
val initialClique = IP.makeIdentity 
val initialSep = IP.makeIdentity 
fun addPotToClique (ip,cp) = IP.multiply ip cp
end (* structure JTI *)

structure JT = JoinTreeFn(structure JTI = JTI) 
	       : JOIN_TREE where type pot = IP.ipot


exception Error of string

type engine = JT.jtree

datatype evidence = Hard of RVar.var * RVar.value
		  | Soft of IP.ipot

fun sendMessage (pot_x, pot_r, pot_y) =
    let
(*	val _ = print "--Sending message\n"
	val _ = print ("from: " ^ (IP.toString pot_x) ^ "\n")
	val _ = print ("sep: " ^ (IP.toString pot_r) ^ "\n")
	val _ = print ("to: " ^ (IP.toString pot_y) ^ "\n")
*)
	val scope_x = RVSet.addList(RVSet.empty, IP.scope pot_x)
	val scope_y = RVSet.addList(RVSet.empty, IP.scope pot_y)
	val scope_r = RVSet.addList(RVSet.empty, IP.scope pot_r)
	val scope_x_sub_r = RVSet.difference(scope_x, scope_r)
	val pot_r' = IP.marginalize pot_x (RVSet.listItems scope_x_sub_r)
	val pot_y' = IP.multiply pot_y (IP.divide pot_r' pot_r)

(*	val _ = print ("pot_r': " ^ (IP.toString pot_r') ^ "\n")
	val _ = print ("result: " ^ (IP.toString pot_y') ^ "\n")
*)
    in
	(pot_r', pot_y')
    end

fun makeConsistent jtree =
let
(*     val _ = print "makeConsistent\n" *)

    fun subTree (JT.Clique (cpot, children)) =
	let
	    (* for a fold across the children *)
	    fun f ((spot, JT.Clique (cpot',grandchildren)), 
		   (new_cpot, trees)) =
		let 
		    (* Message from parent to child *)
		    val (spot1, cpot1') = sendMessage (cpot,spot,cpot')
		    (* Make the child consistant *)
		    val (cpot2', tree) = subTree (JT.Clique (cpot1',grandchildren))
		    (* Message from child to parent *)
		    val (spot2, new_cpot') = sendMessage (cpot2', spot1, 
							      new_cpot)
		in
		    (new_cpot', (spot2,tree)::trees)
		end
		    
	    val (cpot', children') = List.foldl f (cpot,[]) children
	in
	    (cpot', JT.Clique (cpot', children'))
	end
	    
    val (_,jtree') = subTree jtree
in
    jtree'
end

(* Putting it together to build the consistent jtree *)
fun makeEngine (gmodel, _, _) = 
    let
	val ic_jtree = JT.build gmodel
	val c_jtree = makeConsistent ic_jtree
    in
	c_jtree
    end

fun addIPotential jtree ipot = 
    let
	val scope = RVSet.addList (RVSet.empty, IP.scope ipot)
	fun tryAdd (JT.Clique (cpot, sep_subtrees)) =
	    let
		val scope' = RVSet.addList (RVSet.empty, JTI.scope cpot)
	    in
		if RVSet.isSubset(scope, scope') 
		then SOME (JT.Clique (cpot, sep_subtrees))
		else List.foldl (fn (_, SOME x) => SOME x
				  | (child, NONE) => tryAdd child)
				NONE 
				(#2 (ListPair.unzip sep_subtrees))

	    end
    in
	case tryAdd jtree of NONE => raise Error "Cannot add partial pots yet"
			   | SOME x => x
    end

fun query jtree query_vars = 
    let
(* 	val _ = print "query\n" *)

	val query_set = RVSet.addList(RVSet.empty, query_vars)

	val jtree_scope = RVSet.addList(RVSet.empty, JT.scope jtree)

	val _ = if not (RVSet.isSubset(query_set, jtree_scope))
		then raise Error "bad query: not in scope of engine"
		else ()
			
	fun findClique (JT.Clique (cpot, children)) = 
	    let
		val (_,subtrees) = ListPair.unzip children
		val scope = RVSet.addList(RVSet.empty, JTI.scope cpot)
	    in
		if RVSet.isSubset(query_set, scope)
		then let
			val barrens =  RVSet.listItems
					    ( RVSet.difference(scope, 
								query_set))
			val marg = IP.marginalize cpot barrens 
			val (marg,_) = IP.normalize marg

		    in SOME marg end
		else  List.foldr (fn (c,NONE) => findClique c 
    				   | (_,SOME s) => SOME s)
				 NONE subtrees
	    end
	   
	val single_clique = findClique jtree
    in
	case single_clique of
	    SOME cpot => cpot
	  | NONE => outOfCliqueQuery jtree query_vars 
    end

(* TESTME *)
and outOfCliqueQuery jtree query_vars =
    let
(*  	val _ = print "..outOfCliqueQuery size:" *)
(* 	val _ = print ((Int.toString (jtreeSize jtree)) ^ "\n") *)
(* 	val _ = print "query_vars: " *)
(* 	val _ = List.foldl (fn (v,_) => print (" " ^ (RVar.toString v))) *)
(* 			   () query_vars *)
(* 	val _ = print "\n" *)
(* 	val _ = Util.showDot (toDotString jtree) *)

	(* sub-optimal *)
	fun getSubtree (JT.Clique (cpot, sep_subtrees)) query_set =
	    let
		val scope = RVSet.addList(RVSet.empty, 
					   (JTI.scope cpot))
		val missing = RVSet.difference(query_set, scope)

		(* See how each subtree can make up for the lacking vars, 
		 only keeping those which help *)
		val missing_sep_subtrees = 
		    List.map (fn (sep,subtree) => 
				 let val (missing',subtree') 
				       = getSubtree subtree missing
				 in  (missing', sep, subtree') end)
			     sep_subtrees

		(* remove subtrees which are missing the same as we are *)
		val missing_sep_subtrees = 
		    List.filter (fn (d,_,_)=> not (RVSet.equal (d,missing)))
				missing_sep_subtrees

		(* sort the smallest difference first *)
		val missing_sep_subtrees = 
		    ListMergeSort.sort (fn ((d1,_,_), (d2,_,_))=>
					   (RVSet.numItems d1) >
					   (RVSet.numItems d2))
				       missing_sep_subtrees

		(* fold across the subtees, and get a set which
		contains all of the query nodes *)
		fun f ((missing',sep,subtree), (missing, acc)) =
		    if RVSet.numItems missing = 0 
		    then (missing,acc) (* done *)
		    else 
			let val common = RVSet.intersection (missing,
							      missing')
			    val diff = RVSet.difference (missing, missing')
			in
			    if RVSet.numItems diff = 0 
			    then (missing,acc) (* this subtree is no help *)
			    else (common,(sep,subtree)::acc)
			end
		val (missing', sep_subtrees) = List.foldl f (missing,[]) 
							  missing_sep_subtrees
	    in
		(missing', JT.Clique(cpot, sep_subtrees))
	    end

	fun getPots pots (JT.Clique (_, children)) =
	    let
		(* Add potential for each outgoing edge *)
		val pots' = 
		    List.foldl (fn ((sep, JT.Clique (cpot, _)), pots)
				   => (IP.divide cpot sep)::pots)
			       pots children
		(* recurse on children *)
		val pots'' = 
		    List.foldl (fn ((_,subtree), acc) 
				   => getPots acc subtree)
			       pots' children
	    in pots'' end

	val query_set = RVSet.addList(RVSet.empty, query_vars)

(* 	val _ = print "...getSubtree\n" *)

	val (rest, (subtree as (JT.Clique (cpot, _)))) = getSubtree jtree query_set


(* 	val _ = print "subtree size: " *)
(* 	val _ = print ((Int.toString (jtreeSize subtree)) ^ "\n") *)
(* 	val _ = Global.showDot (toDotString subtree) *)

	val _ = if RVSet.numItems rest <> 0 
		then raise Error "Jointree, unknown query vars "
		else ()

(* 	val _ = print "...getPots\n" *)

	val potentials = getPots [cpot] subtree

(* 	val _ = print "...eliminating\n" *)
			 
	val answer = IP.varElim potentials (RVSet.listItems query_set)
(* 	val _ = print "done\n" *)

(* 	val _ = print ("answer : " ^ IP.toString answer) *)
    in
	answer
    end

fun enterEvidence jtree (Hard (var,value) ) = 
    let  (* HACK, represent as soft evidence *)
	val observed_index = RVar.valueToInt value
	val table = List.tabulate (RVar.dim var, fn i => if i = observed_index
							 then 1.0
							 else 0.0)
	val ipot = IP.make([var], table)
    in
	addIPotential jtree ipot
    end
  | enterEvidence jtree (Soft ipot) = addIPotential jtree ipot

end

