functor LazyInferFn(structure GModel : D_G_MODEL
                    structure RVar : D_R_VAR
		    structure IP : D_IPOTENTIAL
		    sharing GModel.RVar = RVar 
		    and GModel.IP = IP ) 
	:> INFERENCE where GModel = GModel
                       and IP = IP
		       and RVar = RVar = 
struct
structure GModel = GModel
structure IP = IP
structure RVar = RVar
structure RVSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)
structure IPSet = ListSetFn(struct 
			    type ord_key = IP.ipot
			    fun compare (ip,ip') = 
				List.collate RVar.compare (IP.scope ip, 
							   IP.scope ip')
			    end)
(* If we already have a potential with the same domain, then mulitply it out *)
fun addPot set pot = 
    case IPSet.find (fn p=> IPSet.Key.compare(p,pot)=EQUAL) set 
     of SOME(p) => IPSet.add (set, IP.multiply p pot)
      | NONE => IPSet.add (set, pot)
fun potsScope pset = IPSet.foldl (fn (ip,vs)=>RVSet.addList(vs,IP.scope ip))
				RVSet.empty pset


structure JTI : JOIN_TREE_INNARDS =
struct
structure GModel = GModel
structure IP = IP
structure RVar = RVar

type pot = IPSet.set

fun scope p =
    let
	val vset = List.foldl (fn (p,s) => RVSet.addList (s,IP.scope p))
			      RVSet.empty (IPSet.listItems p)
    in
	RVSet.listItems vset
    end

fun initialClique v_list = IPSet.empty (*singleton (IP.makeIdentity v_list)*)
fun initialSep v_list = IPSet.singleton (IP.makeIdentity v_list)

fun addPotToClique (pot, pot_set) = addPot pot_set pot
end

structure JT = JoinTreeFn(structure JTI = JTI) 
	       : JOIN_TREE where type pot = IPSet.set


exception Error of string

type engine = JT.jtree

datatype evidence = Hard of RVar.var * RVar.value
		  | Soft of IP.ipot

fun potSetToString pot_set = 
    let
	fun ipotToString ipot = IP.toString ipot
    in
	(IPSet.foldl (fn (p,acc)=> acc ^ " {" 
				   ^ (ipotToString p)
				   ^ " }" )
		     "" pot_set )
    end

fun toDotString jtree =
    let
	val header = "digraph {\n"
	val footer = "}\n"

	fun cliqueToString cpot = "\"" ^ (potSetToString cpot)		  
				  ^ "\""
	fun subTree (JT.Clique (cpot, children)) =
	    let
		val (spots,childs) = ListPair.unzip children
		val child_cpots = List.map (fn (JT.Clique (c,_))=> c) childs
		val clique_string = cliqueToString cpot
		val children_edges =
		    List.foldl (fn (c,acc)=> 
				   acc ^ "\n" ^ clique_string 
				   ^ " -> " ^ (cliqueToString c))
			       "" child_cpots
		val rest = List.foldl (fn (c,acc) => acc ^ (subTree c))
				      "" childs
	    in
		children_edges ^ rest
	    end

	val graph = subTree jtree
    in
	header ^ graph ^ footer
    end

fun marginilize x pots =
    let
	val (marg,rest) = IPSet.partition (fn ip=> Util.member x (IP.scope ip))
					  pots
	val marg_ls = IPSet.listItems marg
	val phi' = List.foldl (fn (p,p') => IP.multiply p p')
			      (hd marg_ls) (tl marg_ls)
	val phi' = IP.marginalize phi' [x]
    in
	addPot rest phi'
    end

fun sendMessage (c_send, sep, other_seps, c_to) =
    let
	val R_s = List.foldl IPSet.union c_send other_seps
	val R_s_scope = potsScope R_s
	val sep_scope = potsScope sep
	val marg_vars = RVSet.difference(R_s_scope, sep_scope)
	val sep' = RVSet.foldl (fn (v,pots) => marginilize v pots)
			       R_s marg_vars
	val c_to' = IPSet.foldl (fn (p,s)=> addPot s p) c_to sep'
		      
(* 	fun printRVSet rvs =  *)
(* 	    RVSet.foldl (fn (v,_) => print (RVar.toString v ^ " ")) *)
(* 			() rvs *)
(* 	val _ = print ("R_s: " ^ (potSetToString R_s) ^ "\n") *)
(* 	val _ = print ("marg_vars: " ) *)
(* 	val _ = printRVSet marg_vars *)
(* 	val _ = print ("\nsep_scope: ") *)
(* 	val _ = printRVSet sep_scope *)
(* 	val _ = print "\n" *)
(* 	val _ = print ("sending: " ^ (potSetToString sep) ^ "\n") *)
(* 	val _ = print ("to: " ^ (potSetToString c_to) ^ "\n") *)
(* 	val _ = print ("is: " ^ (potSetToString message) ^ "\n") *)
    in
	(sep', c_to')
    end

fun collectEvidence jtree =
    let
	fun subTree (JT.Clique (c_i, children)) =
	    let
		(* for a fold across the children *)
		fun f ((sep, jtree), (c_i,acc)) =
		    let
			val jtree' as (JT.Clique(c_j, children')) 
			  = subTree jtree
			val (seps', _ ) =ListPair.unzip children'
			val (sep',c_i') = sendMessage (c_j, sep, seps', c_i)
		    in
			(c_i',
			 (sep', jtree')::acc)
		    end

		val (c_i', children') = List.foldl f (c_i, []) children
	    in
		JT.Clique (c_i', children)
	    end
    in
	subTree jtree
    end

fun distributeEvidence jtree = 
    let
	fun subTree (JT.Clique (c_i, children), in_sep) =
	    let
		val (seps, _) = ListPair.unzip children
		(* for a fold across the children *)
		fun f ((sep, jtree), acc) =
		    let
			val JT.Clique(c_j,children) = jtree
			val seps' = List.filter (fn s=>not (IPSet.equal(sep,s)))
						seps
			val seps' = case in_sep of SOME(s) => s::seps'
						 | NONE => seps'
			val (sep', c_j') = sendMessage (c_i, sep, seps', c_j)
			val jtree = JT.Clique(c_j', children)
			val jtree' = subTree (jtree, SOME sep)

		    in
			(sep', jtree')::acc
		    end
	    in
		JT.Clique (c_i, List.foldl f [] children)
	    end
    in
	subTree (jtree, NONE)
    end

fun makeConsistent jtree =
    distributeEvidence (collectEvidence jtree)

(* Putting it together to build the consistent jtree *)
fun makeEngine (gmodel, _, _) = 
    let
	val ic_jtree = JT.build gmodel
(* 	val _ = Util.showDot (toDotString ic_jtree) *)
(* 	val _ = Posix.Process.sleep (Time.fromReal 1.0) *)
	val c_jtree = makeConsistent ic_jtree
(* 	val _ = Util.showDot (toDotString c_jtree) *)
    in
	c_jtree
    end

fun enterEvidence eng evd = raise Error "unimplemented"

fun query jtree query_vars =
    let
(* 	val _ = print "query_vars: " *)
(* 	val _ = List.foldl (fn (v,_) => print (" " ^ (RVar.toString v))) *)
(* 			   () query_vars *)

	val query_set = RVSet.addList(RVSet.empty, query_vars)

	val jtree_scope = RVSet.addList(RVSet.empty, JT.scope jtree)

	val _ = if not (RVSet.isSubset(query_set, jtree_scope))
		then raise Error "bad query: not in scope of engine"
		else ()
	
	fun findClique (JT.Clique (cpot, children)) = 
	    let
		val (_,subtrees) = ListPair.unzip children
		val scope = RVSet.addList(RVSet.empty, JTI.scope cpot)
	    in
		if RVSet.isSubset(query_set, scope)
		then let
			val result = IP.varElim (IPSet.listItems cpot) 
						(RVSet.listItems query_set)
		    in 
			SOME result 
		    end
		else  List.foldr (fn (c,NONE) => findClique c 
    				   | (_,SOME s) => SOME s)
				 NONE subtrees
	    end
	   
	val single_clique = findClique jtree
    in
	case single_clique of
	    SOME cpot => cpot
	  | NONE => raise Error "out of clique not implemented"
    end


end
