(* D_IPOTENTIAL : Interface for a internal potential over discrete RVs. 
   
   Implementation of a discrite potential by maintaing a mulit-dimenional 
   table.

 *)
signature D_IPOTENTIAL =
sig
    structure RVar : D_R_VAR

    exception Error of string

    type ipot

    val make : (RVar.var list * real list) -> ipot

    (* entries initialised to one -- a empty list is a scalar *)
    val makeIdentity : RVar.var list -> ipot 

    val isSubset : (ipot*ipot) -> bool

    val isApproxEqual : real -> (ipot*ipot) -> bool
		
    (* index : get the value of an entry *)
    val index : ipot -> RVar.value list -> real

    val scope : ipot -> RVar.var list

    val size : ipot -> int

    val multiply : ipot -> ipot -> ipot

    val divide : ipot -> ipot -> ipot

    val marginalize : ipot -> RVar.var list -> ipot

    val sum : ipot -> real

    val scale : ipot -> real -> ipot

    val normalize : ipot -> (ipot * real)

    val toString : ipot -> string

    (* eliminate vars *not* in the list *)
    val varElim : ipot list -> RVar.var list -> ipot
end
