functor DIPotentialFn(structure RVar : D_R_VAR
		       structure Table : TABLE) 
	:> D_IPOTENTIAL where RVar = RVar =
struct


structure RVar = RVar
structure T = Table

structure RVarSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)

exception Error of string

(* Represent a potential as a var list and table. 
  - The vars must always be in increasing order of ids.
*) 
type ipot = RVar.var list * T.table

fun toString (vl,t) = "Potential vars: " ^
		       (List.foldl (fn (v,acc) => acc ^ " " ^(RVar.toString v))
				   "" vl) ^ 
		       "\n " ^ (T.toString t)

(* This is an ugly hack, but seems to work *)
fun reSort (vl,entries) =
    let
	val id_list = List.map RVar.id vl
 	val dim_list = List.map RVar.dim vl
	val vl = ListPair.zip (id_list, dim_list)

	val vl' = ListMergeSort.sort (fn ((id,_), (id',_))=> id > id') vl
	val (ids', dims') = ListPair.unzip vl'
	fun new2old i = 
	    let
		(* expanded index to the old base *)
		fun expand _ [] = [] 
		  | expand n ((id,dim)::vs) = (id, n mod dim) :: 
					      (expand (n div dim) vs)
		(* find the ids value *)
		fun findId id' ((id,v)::rest) =
		    if id'=id then v else findId id' rest
		(* reorder to new base *)
		fun reorder [] _ = []
		  | reorder (id::ids) expanded = findId id expanded :: 
						 (reorder ids expanded)
		(* compress down to old base index *)
		fun compress [] = 0
		  | compress ((dim,v)::rest) = v + dim * (compress rest)

		val expanded = expand i vl'
		val reordered_vals = reorder id_list expanded
		val dim_vals' = ListPair.zip (dim_list, reordered_vals)
	    in
		compress dim_vals'
	    end
	val entries' = List.tabulate (List.length entries, 
				      fn x => List.nth (entries, new2old x))
    in
	(List.map RVar.make vl', entries')
    end


fun make (vl,entries) = 
    let
	val (vl,entries) = reSort (vl, entries)

	val id_list = List.map RVar.id vl
 	val dim_list = List.map RVar.dim vl
	val size = List.foldr op* 1 dim_list
	val _ =	if not (Util.sorted op> id_list) 
		then raise Error "IPotential.make: vlist not sorted"
		else if size
			<> List.length entries 
		then raise Error "IPotential.make: bad spec"
		else ()
    in
	(vl, T.make dim_list entries)
    end

fun makeIdentity vl =
    let 
 	val dim_list = List.map RVar.dim vl
	val size = List.foldr op* 1 dim_list
    in
	make (vl, List.tabulate (size, (fn _=>1.0)))
    end

fun scope (vl,_) = vl

fun size (_,t) = Table.size t

fun isSubset (f,f') =
    let
	val v = RVarSet.addList (RVarSet.empty, scope f)
	val v' = RVarSet.addList (RVarSet.empty, scope f')
    in
	RVarSet.isSubset (v,v')
    end

fun isApproxEqual e ((vl,t),(vl',t')) = 
    (vl = vl') andalso T.isApproxEqual e (t,t')

fun index (vl,table) index =
    let
	val id_list = List.map RVar.id vl
	val dim_list = List.map RVar.dim vl
	val index = List.map RVar.valueToInt index
	val _ = ListPair.foldr (fn (v,d,_)=>if v >= d orelse v < 0 
					    then raise Error "index: bad spec"
					    else ())
			       () (index,dim_list)
    in
	T.getValue index table 
    end

fun reshape (v1,t1) (v2,t2) =
    let
	val v1s = ListPair.zip(List.map RVar.id v1, List.map RVar.dim v1)
	val v2s = ListPair.zip(List.map RVar.id v2, List.map RVar.dim v2)

	(* filter same vars while maintaining thier sorted order *)
	fun newVars ((v,d)::vs) ((v',d')::vs') =
	    (case Int.compare (v,v')
	     of EQUAL => if d <> d' 
			 then raise Error "reshape: inconsistant vars"
			 else (v,d) :: (newVars vs vs')
	      | LESS => (v,d) :: (newVars vs ((v',d')::vs'))
	      | GREATER => (v',d') :: (newVars ((v,d)::vs) vs'))
	  | newVars [] vs = vs
	  | newVars vs [] = vs

	val v's =  newVars v1s v2s 

	val (f1,f2) = List.foldr 
			  (fn ((v,d), (f1,f2)) => 
			      (if Util.member (v,d) v1s then 1::f1 else d::f1, 
			       if Util.member (v,d) v2s then 1::f2 else d::f2))
			  ([],[]) v's

	val new_vars = List.map RVar.make v's
	val new_t1 = T.reshape f1 t1
	val new_t2 = T.reshape f2 t2
    in
	((new_vars, new_t1), 
	 (new_vars, new_t2))
    end

fun multiply ipot1 ipot2 = 
    let
	val ((new_vars, new_t1), (_, new_t2)) = reshape ipot1 ipot2
    in
	(new_vars, T.multiply(new_t1, new_t2))
    end 


fun divide ipot1 ipot2 = 
    let
	val ((new_vars, new_t1), (_, new_t2)) = reshape ipot1 ipot2
    in
	(new_vars, T.divide(new_t1, new_t2))
    end 


fun marginalize ((vl,t) ) ml = 
    let
	fun makeFormat [] = []
	  | makeFormat (v::vs) = if (Util.member v ml) 
				then 0 :: (makeFormat vs)
				 else 1 :: (makeFormat vs)

	val format = makeFormat vl
	val rest = List.filter (fn x=> not (Util.member x ml)) vl
    in
	(rest, T.sumv format t)
    end

fun sum (_,t) = T.sum t

fun scale (vl,t) r = (vl, T.scale t r)

fun normalize f = 
    let
	val z = (sum f)
    in
	(scale f (1.0/z), z)
    end

fun varElim ipots qvars = (* TODO? -- allow the specification of ordering *)
    let
	fun elimVar ipots var = 
	    let 
		val (elim_ipots, rest_ipots) = 
		    List.partition (fn f=> Util.member var (scope f)) ipots
		val ipot' = List.foldr (fn (p,p') => multiply p p')
				      (hd elim_ipots) (tl elim_ipots)
		val ipot' = marginalize ipot' [var]
	    in
		ipot' :: rest_ipots
	    end

	val scope = List.foldr (fn (p,s)=>RVarSet.addList(s, scope p)) 
			       RVarSet.empty ipots

	val evars = RVarSet.difference(scope, RVarSet.addList(RVarSet.empty,qvars))

	val pots = List.foldr (fn (v,ps)=>elimVar ps v)
			    ipots (RVarSet.listItems evars)

	val final_pot = List.foldr (fn (p,p')=>multiply p p')
				   (hd pots) (tl pots)

	val (norm,z) = normalize final_pot
    in
	norm
    end


end
