(* TABLE : A simple multi-dimentional table of reals.

exceptions:
     Error "description" : raised on incorrect usage. (Bad formats,...)

 *)
signature TABLE = 
sig
    type table

    exception Error of string

    (* make : dimensions -> entries -> table *)
     val make : int list -> real list -> table

     val size : table -> int

    (* getValue : indexes -> table -> value *)
     val getValue : int list -> table -> real

    (* reshape : format list -> table -> reshaped table 
      
       - A one in the format list means use the existing dimension.
        Otherwise, a new dimesion (of the specified size) is created,
        and the existing table values are used.  *)

    val reshape : int list -> table -> table

    val sum : table -> real

    val scale : table -> real -> table

    val multiply : table*table -> table (* table sizes must agree *)

    val divide : table*table -> table (* table sizes must agree *)

    (* sumv : format list -> table -> reduced_table 
       - A one in the format list means *keep* this dimension.
         Otherwise, it is summed out.
     *)			      
    val sumv : int list -> table -> table

    val isApproxEqual : real -> (table*table) -> bool

    val toString : table -> string

end
