structure IPotentialTester =
struct

structure IP = GM.IPotential
structure RV = GM.RVar
structure Tester = TesterFn()

val rand = Random.rand (6,9)

fun randomReals 0 = []
  | randomReals n = Real.abs (Random.randReal rand) :: (randomReals (n-1))

fun randomPot vars =
    let
	val size = List.foldl op* 1 (List.map RV.dim vars)
    in
	IP.make (vars, randomReals size)
    end

fun commutativeTest IPop =
    let
	val a = RV.newVar 2
	val b = RV.newVar 3
	val c = RV.newVar 4
	val d = RV.newVar 5

	fun test 0 = true
	  | test n = 
	    let 
		val pot1 = randomPot [a]
		val pot2 = randomPot [a,b]
		val pot3 = randomPot [a,b,c]
		val pot4 = randomPot [b,c]
		val pot5 = randomPot [b,c,d]
		val pot6 = randomPot [c,d]
		val pot7 = randomPot [c,d]
		val opList = List.foldr (fn (x,y)=>IPop x y) (IP.makeIdentity [])
		val ans1 = opList [pot1,pot2,pot3,pot4,pot5,pot6,pot7]
		val ans2 = opList [pot2,pot3,pot4,pot5,pot6,pot7,pot1]
		val ans3 = opList [pot3,pot4,pot5,pot6,pot7,pot1,pot2]
		val ans4 = opList [pot3,pot4,pot6,pot5,pot7,pot1,pot2]
		val ans5 = opList [pot3,pot4,pot5,pot6,pot1,pot7,pot2]
		val ans6 = opList [pot3,pot4,pot1,pot6,pot5,pot7,pot2]
	    in
		IP.isApproxEqual 10e~8 (ans1,ans2) andalso
		IP.isApproxEqual 10e~8 (ans1,ans3) andalso
		IP.isApproxEqual 10e~8 (ans1,ans4) andalso
		IP.isApproxEqual 10e~8 (ans1,ans5) andalso
		IP.isApproxEqual 10e~8 (ans1,ans6) andalso
		test (n-1)
	    end
    in
	fn () => test 5
    end

fun margTest () =
    let
	val a = RV.newVar 2
	val b = RV.newVar 3
	val c = RV.newVar 4
	val d = RV.newVar 5
		
	val pot1 = randomPot [a,b,c,d]
	val pot2 = IP.marginalize pot1 [a,c,d]
	val pot3 = IP.marginalize pot1 [d]
	val pot4 = IP.marginalize pot3 [c]
	val pot5 = IP.marginalize pot4 [a]
	val pot6 = IP.marginalize pot1 [a,d]
	val pot7 = IP.marginalize pot6 [c]
    in
	IP.isApproxEqual 10e~8 (pot2, pot5) andalso
	IP.isApproxEqual 10e~8 (pot2, pot7)
    end

val _ = Tester.setName "IPotential"

val _ = Tester.addTest "test 1 (commutative multiplication)" 
		       (commutativeTest IP.multiply)
val _ = Tester.addTest  "test 2 (marginalize)" 
			(margTest)
val PASSED_ALL_TESTS = Tester.runTests () 

end
