structure Basic : JOINTREE_INNARDS =
struct
structure GModel = GM.GModel
structure IP = GM.IPotential
structure Var = GM.Var
structure VarSet = ListSetFn(struct 
			     type ord_key = Var.var
			     val compare = Var.compare 
			     end)

type clique_pot = IP.ipot
type sep_pot = IP.ipot

val cliqueScope = IP.scope
val sepScope = IP.scope
val initialCliquePot = IP.makeIdentity 
val initialSepPot = IP.makeIdentity 
val marginalizeCliquePot = IP.marginalize
val marginalizeSepPot = IP.marginalize
val normalizeCliquePot = IP.normalize
val dividePot = IP.divide
fun cliquePotToIPot c = c  
fun sepPotToIPot c = c

fun addPotToClique (ip,cp) = IP.multiply ip cp

fun sendMessage (pot_x, pot_r, pot_y) =
    let
	val scope_x = VarSet.addList(VarSet.empty, IP.scope pot_x)
	val scope_y = VarSet.addList(VarSet.empty, IP.scope pot_y)
	val scope_r = VarSet.addList(VarSet.empty, IP.scope pot_r)
	val scope_x_sub_r = VarSet.difference(scope_x, scope_r)
	val pot_r' = IP.marginalize pot_x (VarSet.listItems scope_x_sub_r)
	val pot_y' = IP.multiply pot_y (IP.divide pot_r' pot_r)
    in
	(pot_r', pot_y')
    end
end (* structure Basic : JOINTREE_INNARDS *)
 

structure JoinTreeTester =
struct

structure GModel = GM.GModel
structure LF = GM.LocalFactor
structure Var = GM.Var
structure BasicJT = JoinTree(structure JTI = Basic)


val [a,b,c,d,e,f,g,h] = List.map Var.make [(0,2),(1,2),(2,2),(3,2),(4,2),
					   (5,2),(6,2),(7,2)]
(* These number do not corrispond with the paper!! *)
val fa = LF.makeCPD ([0.5, 0.5],a,[])
val fb = LF.makeCPD ([0.5, 0.4, 0.5, 0.6],b,[a])
val fc = LF.makeCPD ([0.7, 0.2, 0.3, 0.8],c,[a])
val fd = LF.makeCPD ([0.9, 0.1, 0.5, 0.5],d,[b])
val fe = LF.makeCPD ([0.3, 0.7, 0.6, 0.4],e, [c])
val ff = LF.makeCPD ([0.01, 0.01, 0.01, 0.99,
		      0.99, 0.99, 0.99, 0.01], f, [d,e])
val fg = LF.makeCPD ([0.8, 0.1, 0.2, 0.9],g,[c])
val fh = LF.makeCPD ([0.05, 0.95, 0.95, 0.95,
			    0.95, 0.05, 0.05, 0.05], h, [e,g])

val gmodel = GModel.make [fa, fb, fc, fd, fe, ff, fg, fh]
val jtree = BasicJT.make gmodel

end


(* open GM *)
(* structure JT = JoinTreeInfer(structure GModel=GModel *)
(* 			     structure Var = Var *)
(* 			     structure IP = IPotential) *)
(* structure LF = GModel.LocalFactor *)
(* val [a,b,c,d,e,f,g,h] = List.map Var.make [(0,2),(1,2),(2,2),(3,2),(4,2), *)
(* 					   (5,2),(6,2),(7,2)] *)
(* (\* These number do not corrispond with the paper!! *\) *)
(* val fa = LF.makeCPD ([0.5, 0.5],a,[]) *)
(* val fb = LF.makeCPD ([0.5, 0.4, 0.5, 0.6],b,[a]) *)
(* val fc = LF.makeCPD ([0.7, 0.2, 0.3, 0.8],c,[a]) *)
(* val fd = LF.makeCPD ([0.9, 0.1, 0.5, 0.5],d,[b]) *)
(* val fe = LF.makeCPD ([0.3, 0.7, 0.6, 0.4],e, [c]) *)
(* val ff = LF.makeCPD ([0.01, 0.01, 0.01, 0.99, *)
(* 		      0.99, 0.99, 0.99, 0.01], f, [d,e]) *)
(* val fg = LF.makeCPD ([0.8, 0.1, 0.2, 0.9],g,[c]) *)
(* val fh = LF.makeCPD ([0.05, 0.95, 0.95, 0.95, *)
(* 			    0.95, 0.05, 0.05, 0.05], h, [e,g]) *)

(* val gmodel = GModel.make [fa, fb, fc, fd, fe, ff, fg, fh] *)
(* val jtree = JT.gmodelToJtree gmodel *)
(* val _ = print "Initial Join tree:\n" *)
(* val _ = print (JT.toString jtree) *)
(* val collect_jtree = JT.collectEvidence jtree *)
(* val _ = print "Join tree after collecting evidence:\n" *)
(* val _ = print (JT.toString collect_jtree) *)
(* val dist_jtree = JT.distributeEvidence collect_jtree *)
(* val _ = print "Join tree after distributing evidence:\n" *)
(* val _ = print (JT.toString dist_jtree) *)
