signature PARAMS =
sig
    val num_vars : int
    val max_var_dim : int
    val extra_factors : int
end

(* Create a Random GModel *)
functor RandomGModelFn(structure GModel : D_G_MODEL
		       structure Params : PARAMS) 
	 = 
struct
structure GModel = GModel
structure RVar = GModel.RVar
structure IP = GModel.IP
structure LF = GModel.LocalFactor
structure RVSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)

val rand = Random.rand (6,9)

val num_vars = Params.num_vars
val max_var_dim = Params.max_var_dim
val extra_factors = Params.extra_factors

val var_list = List.map (fn id => RVar.make (id, Random.randRange (2, 
								   max_var_dim)
								  rand))
			(List.tabulate (num_vars, (fn x=>x)))
fun randomReals 0 = []
  | randomReals n = Real.abs (Random.randReal rand) :: (randomReals (n-1))

fun randomFactor () =
    let
	val a = Random.randRange(0, num_vars - 1) rand
	val b = Random.randRange(0, num_vars - 1) rand
	val a = if a = b then a+1 mod num_vars else a
	val v = List.nth(var_list, a)
	val u = List.nth(var_list, b)
	val size = List.foldl op* 1 (List.map RVar.dim [v, u])
	val f = LF.Potential (LF.Potential.make (randomReals size, [v,u]))
    in
	f
    end

fun DFS g =
    let
	open Util
	fun search current found =
	    let
		val nbs = IntSet.addList(IntSet.empty,
					 UGraph.getNeighbors g current)
		val unvisited = IntSet.difference(nbs, found)
		val found' = IntSet.add (found, current)
	    in
		List.foldl (fn (c,found) => search c found)
			   found'
			   (IntSet.listItems unvisited)
	    end
    in
	search 0 IntSet.empty
    end

(* Make a random "regular" graph -- this is fairly highly connected *)
fun fullyConnect () = 
    let
	val graph = UGraph.make var_list
	fun add (factors, graph) =
	    let
		val f = randomFactor ()
		val [a,b] = List.map RVar.id (LF.scope f)
		val graph' = UGraph.addEdge graph (a, b)
		val dfs = DFS graph'
	    in
		if Util.IntSet.numItems dfs = num_vars 
		then (f :: factors, graph')
		else add ((f :: factors), graph')
	    end

	fun remove (factors, graph) =
	    let
		val fi = Random.randRange (0, List.length factors - 1) rand
		val f =  List.nth(factors, fi)
		val [a,b] = List.map RVar.id (LF.scope f)
		val graph' = UGraph.removeEdge graph (a, b)
		val dfs = DFS graph'
	    in
		if Util.IntSet.numItems dfs < num_vars 
		then (factors, graph')
		else remove ((List.take (factors,fi)) @ 
			     (List.drop (factors,fi)), 
			     graph')
	    end
	val (factors, _) = remove (add ([], graph))
    in
	factors
    end

	
val factors = fullyConnect ()

fun extraFactors fs 0 = fs
  | extraFactors fs n = 
    let
	val a = Random.randRange(0, num_vars - 1) rand
	val b = Random.randRange(0, num_vars - 1) rand
	val a = if a = b then a+1 mod num_vars else a
	val v = List.nth(var_list, a)
	val u = List.nth(var_list, b)
	val size = List.foldl op* 1 (List.map RVar.dim [v, u])
	val f = LF.Potential (LF.Potential.make (randomReals size, [v,u]))
    in
	extraFactors (f::fs) (n-1)
    end

		
val factors = extraFactors factors extra_factors						

val model = GModel.make factors
end
