(* DGRAPH : Directed Graph interface. 
 - Vertices are referenced by integer index (starting at zero) 
 - Edges are pairs of vertex indexes (v_in, v_out)
 *)
signature DGRAPH =
sig

    exception Error of string

    type 'a dgraph

    val make : 'a list -> 'a dgraph

    val numVertices : 'a dgraph -> int 
    val addVertex : 'a dgraph -> 'a -> ('a dgraph*int)
    val getVertices : 'a dgraph -> 'a list
    val getVertex : 'a dgraph -> int -> 'a
    val getVertexId : ''a dgraph -> ''a -> int option
    val removeVertex : 'a dgraph -> int -> 'a dgraph
    val disconnectVertex : 'a dgraph -> int -> 'a dgraph
    val updateVertex : 'a dgraph -> int -> 'a -> 'a dgraph
					
    val edgeExists : 'a dgraph -> (int*int) -> bool
    val addEdge : 'a dgraph -> (int*int) -> 'a dgraph
    val getEdges : 'a dgraph -> (int*int) list
    val removeEdge : 'a dgraph -> (int*int) -> 'a dgraph

    val addParents : 'a dgraph -> (int* int list) -> 'a dgraph

    val getChildren : 'a dgraph -> int -> int list
    val getParents : 'a dgraph -> int -> int list 

    val reachable : 'a dgraph -> (int*int) -> int list option

    val toString : 'a dgraph -> ('a -> string) -> string
    val toDotString : 'a dgraph -> ('a -> string) -> string

end
