(* UGRAPH : Undirected Graph interface. 
 - Vertices are referenced by integer index (starting at zero) 
 - Edges are pairs of vertex indexes (v_1, v_1)
 *)
signature UGRAPH =
sig
    structure DGraph : DGRAPH

    type 'a ugraph

    val make : 'a list -> 'a ugraph
			
    val fromDGraph : 'a DGraph.dgraph -> 'a ugraph
       
    val numVertices : 'a ugraph -> int 
    val addVertex : 'a ugraph -> 'a -> ('a ugraph*int)
    val getVertices : 'a ugraph -> 'a list
    val getVertex : 'a ugraph -> int -> 'a
    val getVertexId : ''a ugraph -> ''a -> int option
    val removeVertex : 'a ugraph -> int -> 'a ugraph
    val disconnectVertex : 'a ugraph -> int -> 'a ugraph
    val updateVertex : 'a ugraph -> int -> 'a -> 'a ugraph

    val edgeExists : 'a ugraph -> (int*int) -> bool
    val addEdge : 'a ugraph -> (int*int) -> 'a ugraph
    val connect : 'a ugraph -> int list -> 'a ugraph
    (* NOTE: both (a,b) and (b,a) are returned *)
    val getEdges : 'a ugraph -> (int*int) list 
    val removeEdge : 'a ugraph -> (int*int) -> 'a ugraph

    val getNeighbors : 'a ugraph -> int -> int list

    val moralize : 'a DGraph.dgraph -> 'a ugraph 

    (* given a graph and a weight function, find the max spanning tree *)
    val maxSpanningTree : 'a ugraph -> (('a * 'a) -> int) -> 'a ugraph

    val fullyConnect : 'a ugraph -> 'a ugraph

    val toString : 'a ugraph -> ('a -> string) -> string
    val toDotString : 'a ugraph -> ('a -> string) -> string
end
